import os

from faker import Faker

fake = Faker("en-GB")


def sanitize_name(value):
    return fake.name()


def sanitize_first_name(value):
    return fake.first_name()


def sanitize_last_name(value):
    return fake.last_name()


def sanitize_email(value):
    return fake.unique.email()


def sanitize_company_name(value):
    return fake.unique.company()


def sanitize_phone_number(value):
    return "+44" + fake.msisdn()[3:]


def sanitize_address(value):
    return fake.address().replace("\n", ", ")


def sanitize_website(value):
    return fake.domain_name(2)


def sanitize_text(value):
    if not value:
        return value
    return fake.paragraph(nb_sentences=5)


def sanitize_short_text(value):
    if not value:
        return value
    return fake.text(max_nb_chars=100)


def sanitize_street_address(value):
    return fake.street_address()


def sanitize_city(value):
    return fake.city()


def sanitize_postcode(value):
    return fake.postcode()


def sanitize_eori_number(value):
    return "GB" + str(fake.random_number(digits=12))

def sanitize_ni_eori_number(value):
    return "XI" + str(fake.random_number(digits=12))

def sanitize_eu_eori_number(value):
    eu_countries = ["LU","BE", "FI", "FR", "CZ"]
    return fake.random_element(eu_countries) + str(fake.random_number(digits=5))



def sanitize_sic_number(value):
    return str(fake.random_number(digits=5))


def sanitize_vat_number(value):
    return "GB" + str(fake.random_number(digits=9))


def sanitize_registration_number(value):
    return str(fake.random_number(digits=8))


def sanitize_filename(value):
    return os.path.split(
        fake.file_path(extension=["pdf", "jpg", "csv", "txt", "docx", "ods", "xlsx"])
    )[-1]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest
from six.moves.urllib import parse as urlparse

from ..utils.mysql import (
    decode_mysql_literal,
    decode_mysql_string_literal,
    get_mysqldump_args_and_env_from_url,
    unescape_single_character,
)


@pytest.mark.parametrize(
    "url",
    (
        "mysql://test:test@localhost/test",
        "mysql://localhost:1234/test",
        "mysql://localhost",
    ),
)
def test_get_mysqldump_args_and_env_from_url(url):
    parsed_url = urlparse.urlparse(url)

    if not parsed_url.path:
        with pytest.raises(ValueError):
            get_mysqldump_args_and_env_from_url(url=parsed_url)
        return

    args, env = get_mysqldump_args_and_env_from_url(url=parsed_url)

    assert isinstance(args, list)
    assert isinstance(env, dict)

    assert len(args) > 0
    assert "--complete-insert" in args
    assert "--extended-insert" in args
    assert "--net_buffer_length=10240" in args
    assert args[-1] == parsed_url.path[1:]

    if parsed_url.username:
        index = args.index("-u")
        assert args[index + 1] == parsed_url.username

    if parsed_url.password:
        assert env["MYSQL_PWD"] == parsed_url.password


@pytest.mark.parametrize(
    "text,expected_value",
    (
        ("NULL", None),
        ("TRUE", True),
        ("FALSE", False),
        ("12", 12),
        ("12.5", 12.5),
        ("'test'", "test"),
    ),
)
def test_decode_mysql_literal(text, expected_value):
    assert decode_mysql_literal(text) == expected_value


def test_decode_mysql_literal_invalid_input():
    with pytest.raises(ValueError):
        decode_mysql_literal("ERROR")


@pytest.mark.parametrize(
    "text,expected_output",
    (
        ("'test'", "test"),
        ("'test\\ntest'", "test\ntest"),
        ("'\\0'", "\000"),
        ("'foo", None),
        ("foo'", None),
        ("foo", None),
    ),
)
def test_decode_mysql_string_literal(text, expected_output):
    if expected_output is None:
        with pytest.raises(AssertionError):
            decode_mysql_string_literal(text)
    else:
        assert decode_mysql_string_literal(text) == expected_output


@pytest.mark.parametrize(
    "text,expected_output",
    (
        ("\\\\", "\\"),
        ("\\n", "\n"),
        ("\\r", "\r"),
        ("\\0", "\000"),
        ("\\Z", "\032"),
        ("\\'", "'"),
        ('\\"', '"'),
    ),
)
def test_unescape_single_character(text, expected_output):
    class MockRegexpMatch(object):

        def __init__(self, text):
            self.text = text

        def group(self, index):
            assert index == 0
            return self.text

    assert unescape_single_character(MockRegexpMatch(text)) == expected_output



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import io

import pytest
from six.moves.urllib import parse as urlparse

from ..config import Configuration
from ..dump.mysql import (
    parse_column_names,
    parse_values,
    sanitize,
    sanitize_from_stream,
)

MOCK_MYSQLDUMP_OUTPUT = b"""
--- Fake MySQL database dump

DROP TABLE IF EXISTS `test`;

CREATE TABLE `test` (
`id` int(11) NOT NULL AUTO_INCREMENT,
`created_at` date NOT NULL,
`notes` varchar(255) NOT NULL,
PRIMARY KEY (`id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4;

INSERT INTO `test` (`id`, `created_at`, `notes`) VALUES \
(1,'2018-01-01','Test data 1'),\
(2,'2018-01-02','Test data 2'),\
(3,'2018-01-03','Test data 3');

--- Final line after `INSERT INTO` statement.
"""

MOCK_MYSQLDUMP_OUTPUT_WITH_U2028 = b"""
--- Fake MySQL database dump

DROP TABLE IF EXISTS `test`;

CREATE TABLE `test` (
`id` int(11) NOT NULL AUTO_INCREMENT,
`created_at` date NOT NULL,
`notes` varchar(255) NOT NULL,
PRIMARY KEY (`id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4;

INSERT INTO `test` (`id`, `created_at`, `notes`) VALUES \
(1,'2018-01-01','Test \xe2\x80\xa8 data 1'),\
(2,'2018-01-02','Test data 2'),\
(3,'2018-01-03','Test data 3');

--- Final line after `INSERT INTO` statement.
"""


INVALID_MOCK_MYSQLDUMP_OUTPUT = b"""
--- Fake MySQL database dump

DROP TABLE IF EXISTS `test`;

INSERT INTO `test` (`id`, `created_at`, `notes`) VALUES (1),(2),(3);

--- Final line after `INSERT INTO` statement.
"""


def test_sanitize_wrong_scheme():
    url = urlparse.urlparse("http://localhost/test")
    with pytest.raises(ValueError):
        list(sanitize(url, None))


def test_sanitize_from_stream():
    stream = io.BytesIO(MOCK_MYSQLDUMP_OUTPUT)
    config = Configuration()
    config.sanitizers["test.notes"] = lambda value: "Sanitized"
    dump_output_lines = list(sanitize_from_stream(stream, config))

    assert "--- Fake MySQL database dump" in dump_output_lines
    assert "--- Final line after `INSERT INTO` statement." in dump_output_lines
    assert """INSERT INTO `test` (`id`, `created_at`, `notes`) VALUES \
(1,'2018-01-01','Sanitized'),\
(2,'2018-01-02','Sanitized'),\
(3,'2018-01-03','Sanitized');\
""" in dump_output_lines

def test_sanitize_with_u2028_from_stream():
    stream = io.BytesIO(MOCK_MYSQLDUMP_OUTPUT_WITH_U2028)
    config = Configuration()
    config.sanitizers["test.notes"] = lambda value: "Sanitized"
    dump_output_lines = list(sanitize_from_stream(stream, config))

    assert "--- Fake MySQL database dump" in dump_output_lines
    assert "--- Final line after `INSERT INTO` statement." in dump_output_lines
    assert """INSERT INTO `test` (`id`, `created_at`, `notes`) VALUES \
(1,'2018-01-01','Sanitized'),\
(2,'2018-01-02','Sanitized'),\
(3,'2018-01-03','Sanitized');\
""" in dump_output_lines


def test_skip_table_rows():
    stream = io.BytesIO(MOCK_MYSQLDUMP_OUTPUT)
    config = Configuration()
    config.skip_rows_for_tables.append('test')

    output = list(sanitize_from_stream(stream, config))

    assert output == [
        '',
        '--- Fake MySQL database dump',
        '',
        'DROP TABLE IF EXISTS `test`;',
        '',
        'CREATE TABLE `test` (',
        '`id` int(11) NOT NULL AUTO_INCREMENT,',
        '`created_at` date NOT NULL,',
        '`notes` varchar(255) NOT NULL,',
        'PRIMARY KEY (`id`)',
        ') ENGINE=InnoDB DEFAULT CHARSET=utf8mb4;',
        '',
        '',
        '--- Final line after `INSERT INTO` statement.',
    ]


def test_sanitizer_invalid_input():
    stream = io.BytesIO(INVALID_MOCK_MYSQLDUMP_OUTPUT)
    config = Configuration()
    config.sanitizers["test.notes"] = lambda value: "Sanitized"

    with pytest.raises(ValueError):
        list(sanitize_from_stream(stream, config))


@pytest.mark.parametrize(
    "text,expected_column_names",
    (
        ("`test`", ("test",)),
        ("`test`, `test`", ("test", "test")),
        ("`test`,`test`", ("test", "test")),
    ),
)
def test_parse_column_names(text, expected_column_names):
    assert parse_column_names(text) == expected_column_names


@pytest.mark.parametrize(
    "text,expected_values",
    (
        ("('test'),('test')", (("test",), ("test",))),
        ("(1,2),(3,4),", ((1, 2), (3, 4))),
        ("(TRUE),(FALSE),(NULL)", ((True,), (False,), (None,))),
        ("(x')", ()),  # Invalid data
    ),
)
def test_parse_values(text, expected_values):
    assert tuple(parse_values(text)) == expected_values


@pytest.mark.parametrize('config_type', [
    'no-config', 'empty-config', 'single-column-config'])
@pytest.mark.parametrize('data_label', ['ok', 'invalid'])
def test_optimizations(config_type, data_label):
    if config_type == 'no-config':
        config = None
        decoder_call_count = 0
    else:
        config = Configuration()
        if config_type == 'empty-config':
            decoder_call_count = 0
        else:
            assert config_type == 'single-column-config'
            config.sanitizers["test.notes"] = (lambda x: x)
            decoder_call_count = 3  # Number of rows in test table

    data = {
        'ok': MOCK_MYSQLDUMP_OUTPUT,
        'invalid': INVALID_MOCK_MYSQLDUMP_OUTPUT,
    }[data_label]

    should_raise = (
        config_type == 'single-column-config'
        and data_label == 'invalid')

    dump_stream = io.BytesIO(data)
    if should_raise:
        with pytest.raises(ValueError):
            list(sanitize_from_stream(dump_stream, config))
    else:
        expected_output = data.decode('utf-8').splitlines()
        result = list(sanitize_from_stream(dump_stream, config))
        assert result == expected_output



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import codecs
import re
import subprocess
import io

from ..utils.mysql import (
    decode_mysql_literal,
    encode_mysql_literal,
    get_mysqldump_args_and_env_from_url,
)
from ..config import MYSQLDUMP_DEFAULT_PARAMETERS

#: Regular expression which matches `INSERT INTO` statements produced by the
#: `mysqldump` utility, even when extended inserts have been enabled.
INSERT_INTO_PATTERN = re.compile(
    r"^INSERT INTO `(?P<table>[^`]*)`"
    r" \((?P<columns>.*)\)"
    r" VALUES (?P<values>.*);$"
)


#: Regular expression which matches various kinds of MySQL literals.
VALUE_PATTERN = re.compile(
    r"""
    # Group 1:
    (
        '(?:[^']|''|\\')*(?<![\\])'     # String literal
        |                               # or...
        [^',()]+                        # NULL, TRUE, etc.
    )
    # Group 2:
    (
        [,)]                            # Comma or closing parenthesis.
    )
    """,
    re.VERBOSE,
)


def sanitize(url, config):
    """
    Obtains dump of MySQL database by executing `mysqldump` command and
    sanitizes it output.

    :param url: URL to the database which is going to be sanitized, parsed by
                Python's URL parser.
    :type url: urllib.urlparse.ParseResult

    :param config: Optional sanitizer configuration to be used for sanitation
                   of the values stored in the database.
    :type config: database_sanitizer.config.Configuration|None
    """
    if url.scheme != "mysql":
        raise ValueError("Unsupported database type: '%s'" % (url.scheme,))

    args, env = get_mysqldump_args_and_env_from_url(url=url)

    extra_params = MYSQLDUMP_DEFAULT_PARAMETERS
    if config:
        extra_params = config.mysqldump_params

    process = subprocess.Popen(
        args=["mysqldump"] + args + extra_params,
        env=env,
        stdout=subprocess.PIPE,
    )

    return sanitize_from_stream(stream=process.stdout, config=config)


def sanitize_from_stream(stream, config):
    """
    Reads dump of MySQL database from given stream and sanitizes it.

    :param stream: Stream where the database dump is expected to be available
                   from, such as stdout of `mysqldump` process.
    :type stream: file

    :param config: Optional sanitizer configuration to be used for sanitation
                   of the values stored in the database.
    :type config: database_sanitizer.config.Configuration|None
    """
    for line in io.TextIOWrapper(stream, encoding="utf-8"):
        # Eat the trailing new line.
        line = line.rstrip("\n")

        # If there is no configuration it means that there are no sanitizers
        # available.
        if not config:
            yield line
            continue

        # Does the line contain `INSERT INTO` statement? If not, use the line
        # as-is and continue into next one.
        insert_into_match = INSERT_INTO_PATTERN.match(line)
        if not insert_into_match:
            yield line
            continue

        table_name = insert_into_match.group("table")
        column_names = parse_column_names(insert_into_match.group("columns"))

        # Skip `INSERT INTO` statement if table rows are configured
        # to be skipped.
        if table_name in config.skip_rows_for_tables:
            continue

        # Collect sanitizers possibly used for this table and place them into
        # a dictionary from which we can look them up by index later.
        sanitizers = {}
        for index, column_name in enumerate(column_names):
            sanitizer = config.get_sanitizer_for(
                table_name=table_name,
                column_name=column_name,
            )
            if sanitizer:
                sanitizers[index] = sanitizer

        # If this table has no sanitizers available, use the line as-is and
        # continue into next line.
        if len(sanitizers) == 0:
            yield line
            continue

        # Constructs list of tuples containing sanitized column names.
        sanitized_value_tuples = []
        for values in parse_values(insert_into_match.group("values")):
            if len(column_names) != len(values):
                raise ValueError("Mismatch between column names and values")
            sanitized_values = []
            for index, value in enumerate(values):
                sanitizer_callback = sanitizers.get(index)
                if sanitizer_callback:
                    value = sanitizer_callback(value)
                sanitized_values.append(encode_mysql_literal(value))
            sanitized_value_tuples.append(sanitized_values)

        # Finally create new `INSERT INTO` statement from the sanitized values.
        yield "INSERT INTO `%s` (%s) VALUES %s;" % (
            table_name,
            ", ".join("`" + column_name + "`" for column_name in column_names),
            ",".join(
                "(" + ",".join(value_tuple) + ")"
                for value_tuple in sanitized_value_tuples
            ),
        )


def parse_column_names(text):
    """
    Extracts column names from a string containing quoted and comma separated
    column names of a table.

    :param text: Line extracted from MySQL's `INSERT INTO` statement containing
                 quoted and comma separated column names.
    :type text: str

    :return: Tuple containing just the column names.
    :rtype: tuple[str]
    """
    return tuple(
        re.sub(r"^`(.*)`$", r"\1", column_data.strip())
        for column_data in text.split(",")
    )


def parse_values(text):
    """
    Parses values from a string containing values from extended format `INSERT
    INTO` statement. Values will be yielded from the function as tuples, with
    one tuple per row in the table.

    :param text: Text extracted from MySQL's `INSERT INTO` statement containing
                 quoted and comma separated column values.
    :type text: str
    """
    assert text.startswith("(")
    pos = 1
    values = []
    text_len = len(text)
    while pos < text_len:
        match = VALUE_PATTERN.match(text, pos)
        if not match:
            break
        value = match.group(1)
        values.append(decode_mysql_literal(value.strip()))
        pos += len(value) + 1
        if match.group(2) == ")":
            # Skip comma and open parenthesis ",("
            pos += 2
            yield tuple(values)
            values = []



import os
import logging
from datetime import datetime

import boto3
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from django_db_anonymiser.database_sanitizer.config import Configuration
from django_db_anonymiser.database_sanitizer.dump import run

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--keep-local-dumpfile",
            action="store_true",
            help="Keep local dump file, rather than cleaning it up.",
        )
        parser.add_argument(
            "--skip-s3-upload",
            action="store_true",
            help="Skip uploading to S3.",
        )
        parser.add_argument(
            "--presign",
            action="store_true",
            help="Generates and logs a presigned URL for the uploaded file.",
        )
        parser.add_argument(
            "--add-timestamp",
            action="store_true",
            help="Add current timestamp to dump file name.",
        )

    def configure(self):
        self.keep_local_dumpfile = False
        self.skip_s3_upload = False
        self.presign = False
        base_dump_file_name = settings.DB_ANONYMISER_DUMP_FILE_NAME
        if self.add_timestamp:
            now = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
            base_dump_file_name = f"{now}-{base_dump_file_name}"
        dump_file_name = base_dump_file_name
        if getattr(settings, "DB_ANONYMISER_AWS_STORAGE_KEY", None):
            dump_file_name = (
                f"{settings.DB_ANONYMISER_AWS_STORAGE_KEY}/{dump_file_name}"
            )
        self.dump_file_name = dump_file_name
        self.temporary_dump_location = getattr(
            settings,
            "DB_ANONYMISER_TEMPORARY_DUMP_LOCATION",
            f"/tmp/{base_dump_file_name}",
        )
        try:
            self.config_location = settings.DB_ANONYMISER_CONFIG_LOCATION
        except AttributeError:
            raise CommandError(
                "DB_ANONYMISER_CONFIG_LOCATION must be set in django settings."
            )
        additional_s3_params = {}
        aws_endpoint_url = getattr(settings, "DB_ANONYMISER_AWS_ENDPOINT_URL", None)
        if aws_endpoint_url:
            additional_s3_params[
                "endpoint_url"
            ] = settings.DB_ANONYMISER_AWS_ENDPOINT_URL
        self.s3_client = boto3.client(
            "s3",
            aws_access_key_id=settings.DB_ANONYMISER_AWS_ACCESS_KEY_ID,
            aws_secret_access_key=settings.DB_ANONYMISER_AWS_SECRET_ACCESS_KEY,
            region_name=settings.DB_ANONYMISER_AWS_REGION,
            **additional_s3_params,
        )
        self.s3_bucket_name = settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME

    def handle(self, *args, **options):
        logger.info("Starting DB dump and anonymiser")

        self.add_timestamp = False
        if options["add_timestamp"]:
            self.add_timestamp = True

        self.configure()

        if options["keep_local_dumpfile"]:
            self.keep_local_dumpfile = True

        if options["skip_s3_upload"]:
            self.skip_s3_upload = True

        if options["presign"]:
            self.presign = True

        try:
            self.dump_anonymised_db()
            self.write_to_s3()
            logger.info("DB dump and anonymiser was successful!")
            if self.presign:
                self.generate_presigned_url()
        finally:
            self.cleanup()

    def dump_anonymised_db(self):
        db_details = settings.DATABASES["default"]
        postgres_url = f"postgresql://{db_details['USER']}:{db_details['PASSWORD']}@{db_details['HOST']}:{db_details['PORT']}/{db_details['NAME']}"
        logger.info(
            "Writing anonymised dumpfile to temporary location %s", self.dump_file_name
        )
        with open(self.temporary_dump_location, "w") as outfile:
            run(
                url=postgres_url,
                config=Configuration.from_file(self.config_location),
                output=outfile,
            )
        logger.info("Writing anonymised dumpfile complete")

    def write_to_s3(self):
        if self.skip_s3_upload:
            return
        logger.info("Writing file %s to S3", self.dump_file_name)
        self.s3_client.upload_file(
            self.temporary_dump_location, self.s3_bucket_name, self.dump_file_name
        )
        logger.info("Writing file to S3 complete")

    def generate_presigned_url(self):
        presigned = self.s3_client.generate_presigned_url(
            "get_object",
            Params={"Bucket": self.s3_bucket_name, "Key": self.dump_file_name},
            ExpiresIn=600,
        )
        logger.info("Presigned URL: %s", presigned)

    def cleanup(self):
        if self.keep_local_dumpfile:
            return
        logger.info("Cleaning up temporary files")
        try:
            os.remove(self.temporary_dump_location)
        except FileNotFoundError:
            pass
        logger.info("Clean up complete")



import os
from datetime import datetime
from unittest.mock import ANY
from unittest.mock import patch

from django.conf import settings
from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import override_settings, TransactionTestCase

import boto3
import pytest
import requests
from moto import mock_aws


@pytest.mark.skipif(
    os.getenv("CIRCLECI") != "true",
    reason="Skipped because test requires real postgres db.",
)
@mock_aws
class TestDumpAndAnonmyiseCommand(TransactionTestCase):
    def setUp(self):
        self.aws = boto3.client("s3", region_name=settings.DB_ANONYMISER_AWS_REGION)
        self.aws.create_bucket(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME,
            CreateBucketConfiguration={
                "LocationConstraint": settings.DB_ANONYMISER_AWS_REGION
            },
        )

    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.run"
    )
    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.Configuration"
    )
    def test_dump_and_anonymise_calls_anonymiser(
        self, mocked_configuration, mocked_anonymiser_run
    ):
        call_command(
            "dump_and_anonymise", keep_local_dumpfile=False, skip_s3_upload=True
        )
        call_args, call_kwargs = mocked_anonymiser_run.call_args
        assert (
            call_kwargs["url"]
            == f"postgresql://{settings.DATABASES['default']['USER']}:{settings.DATABASES['default']['PASSWORD']}@{settings.DATABASES['default']['HOST']}:{settings.DATABASES['default']['PORT']}/{settings.DATABASES['default']['NAME']}"
        )
        assert call_kwargs["config"] == mocked_configuration.from_file.return_value
        assert call_kwargs["output"].name == "/tmp/anonymised.sql"
        # Ensure skip_s3_upload was respected
        bucket_contents = self.aws.list_objects(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME
        ).get("Contents", [])
        assert bucket_contents == []

    def test_dump_and_anonymise_writes_sql(self):
        user = User.objects.create(
            first_name="Bob",
            last_name="Benson",
            email="bob.benson@example.net",  # /PS-IGNORE
            username="bob.benson",
        )
        call_command(
            "dump_and_anonymise", keep_local_dumpfile=True, skip_s3_upload=True
        )
        dump_sql = ""
        with open(f"/tmp/{settings.DB_ANONYMISER_DUMP_FILE_NAME}") as f:
            dump_sql = f.read()
        assert user.username in dump_sql
        assert user.first_name not in dump_sql
        assert user.last_name not in dump_sql
        assert user.email not in dump_sql

    def test_dump_and_anonymise_writes_to_s3(self):
        call_command("dump_and_anonymise", keep_local_dumpfile=False)
        bucket_contents = self.aws.list_objects(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME
        ).get("Contents", [])
        assert bucket_contents[0]["Key"] == settings.DB_ANONYMISER_DUMP_FILE_NAME

    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.os.remove"
    )
    def test_dump_and_anonymise_clears_local_file(self, mocked_os_remove):
        call_command("dump_and_anonymise")
        mocked_os_remove.assert_called_with(
            f"/tmp/{settings.DB_ANONYMISER_DUMP_FILE_NAME}"
        )

    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.os.remove"
    )
    def test_dump_and_anonymise_keeps_local_file(self, mocked_os_remove):
        call_command("dump_and_anonymise", keep_local_dumpfile=True)
        assert not mocked_os_remove.called

    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.logger"
    )
    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.Command.generate_presigned_url"
    )
    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.os.remove"
    )
    def test_dump_and_anonymise_no_arguments(
        self, mocked_os_remove, mocked_presign, mocked_logger
    ):
        call_command("dump_and_anonymise")
        bucket_contents = self.aws.list_objects(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME
        ).get("Contents", [])
        assert bucket_contents[0]["Key"] == settings.DB_ANONYMISER_DUMP_FILE_NAME
        mocked_os_remove.assert_called_with(
            f"/tmp/{settings.DB_ANONYMISER_DUMP_FILE_NAME}"
        )
        mocked_presign.assert_not_called()
        mocked_logger.info.assert_any_call("DB dump and anonymiser was successful!")
        mocked_logger.info.assert_any_call(
            "Writing anonymised dumpfile to temporary location %s",
            settings.DB_ANONYMISER_DUMP_FILE_NAME,
        )
        mocked_logger.info.assert_any_call("Writing file to S3 complete")

    @patch(
        "django_db_anonymiser.db_anonymiser.management.commands.dump_and_anonymise.logger"
    )
    def test_dump_and_anonymise_with_presign(self, mocked_logger):
        call_command("dump_and_anonymise", presign=True)
        bucket_contents = self.aws.list_objects(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME
        ).get("Contents", [])
        assert bucket_contents[0]["Key"] == settings.DB_ANONYMISER_DUMP_FILE_NAME
        mocked_logger.info.assert_any_call("Presigned URL: %s", ANY)
        presigned_url = mocked_logger.info.call_args_list[6][0][1]
        assert requests.get(presigned_url).status_code == 200

    def test_dump_and_anonymise_with_timestamp(self):
        now = datetime.now()
        patch(
            "django_db_anonymiser.db_anonymiser.management.commands.datetime.datetime.now",
            return_value=now,
        )
        call_command("dump_and_anonymise", add_timestamp=True)
        bucket_contents = self.aws.list_objects(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME
        ).get("Contents", [])
        assert (
            bucket_contents[0]["Key"]
            == f"{now.strftime('%Y-%m-%d-%H:%M:%S')}-{settings.DB_ANONYMISER_DUMP_FILE_NAME}"
        )

    @override_settings(DB_ANONYMISER_AWS_STORAGE_KEY="test")
    def test_dump_and_anonymise_with_s3_key(self):
        call_command("dump_and_anonymise")
        bucket_contents = self.aws.list_objects(
            Bucket=settings.DB_ANONYMISER_AWS_STORAGE_BUCKET_NAME
        ).get("Contents", [])
        assert (
            bucket_contents[0]["Key"] == f"test/{settings.DB_ANONYMISER_DUMP_FILE_NAME}"
        )



import pytest


@pytest.fixture
def mock_writer(mocker):
    """Mock writer.

    Returns a mock writer object with info, success, notice, warning, and error
    methods.
    """
    writer = mocker.Mock()
    writer.info = mocker.Mock()
    writer.success = mocker.Mock()
    writer.notice = mocker.Mock()
    writer.warning = mocker.Mock()
    writer.error = mocker.Mock()
    return writer



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import io
from collections import namedtuple

import mock
import pytest
from six.moves.urllib import parse as urlparse

from ..config import Configuration
from ..dump import postgres as dump_postgres
from ..dump.postgres import parse_column_names, parse_values, sanitize
from ..utils.postgres import decode_copy_value

MOCK_PG_DUMP_OUTPUT = b"""
--- Fake PostgreSQL database dump

COMMENT ON SCHEMA "public" IS 'standard public schema';

CREATE TABLE "public"."test" (
"id" integer NOT NULL,
"created_at" timestamp with time zone NOT NULL,
"notes" character varying(255) NOT NULL
);

COPY "public"."test" ("id", "created_at", "notes") FROM stdin;
1\t2018-01-01 00:00:00\tTest data 1
2\t2018-01-02 00:00:00\tTest data 2
3\t2018-01-03 00:00:00\tTest data 3
\\.

--- Final line after `COPY` statement
""".strip()


INVALID_MOCK_PG_DUMP_OUTPUT = b"""
--- Fake PostgreSQL database dump

COMMENT ON SCHEMA "public" IS 'standard public schema';

COPY "public"."test" ("id", "created_at", "notes") FROM stdin;
1\t2018-01-01 00:00:00 Test data 1
2\t2018-01-02 00:00:00 Test data 2
3\t2018-01-03 00:00:00 Test data 3
\\.

--- Final line after `COPY` statement
""".strip()


def create_mock_popen(mock_pg_dump_output):
    def mock_popen(cmd_args, stdout):
        mock_pipe_type = namedtuple("mock_pipe", ("stdout",))
        mock_stdout = io.BytesIO(mock_pg_dump_output)
        return mock_pipe_type(stdout=mock_stdout)
    return mock_popen


def test_sanitize():
    url = urlparse.urlparse("postgres://localhost/test")
    config = Configuration()
    config.sanitizers["test.notes"] = lambda value: "Sanitized"

    with mock.patch("subprocess.Popen", side_effect=create_mock_popen(MOCK_PG_DUMP_OUTPUT)):
        dump_output_lines = list(sanitize(url, config))

    assert "--- Fake PostgreSQL database dump" in dump_output_lines
    assert "--- Final line after `COPY` statement" in dump_output_lines
    assert "2\t2018-01-02 00:00:00\tSanitized" in dump_output_lines


def test_skip_table_rows():
    url = urlparse.urlparse("postgres://localhost/test")
    config = Configuration()
    config.skip_rows_for_tables.append('test')

    with mock.patch("subprocess.Popen",
                    side_effect=create_mock_popen(MOCK_PG_DUMP_OUTPUT)):
        output = list(sanitize(url, config))

    assert output == [
        '--- Fake PostgreSQL database dump',
        '',
        'COMMENT ON SCHEMA "public" IS \'standard public schema\';',
        '',
        'CREATE TABLE "public"."test" (',
        '"id" integer NOT NULL,',
        '"created_at" timestamp with time zone NOT NULL,',
        '"notes" character varying(255) NOT NULL',
        ');',
        '',
        '',
        '--- Final line after `COPY` statement'
    ]


def test_sanitizer_invalid_input():
    url = urlparse.urlparse("postgres://localhost/test")

    config = Configuration()
    config.sanitizers["test.notes"] = lambda value: "Sanitized"

    with mock.patch("subprocess.Popen", side_effect=create_mock_popen(INVALID_MOCK_PG_DUMP_OUTPUT)):
        with pytest.raises(ValueError):
            # Yes, we need the list() function there to eat the yields.
            list(sanitize(url, config))


def test_sanitizer_invalid_scheme():
    url = urlparse.urlparse("http://localhost/test")
    with pytest.raises(ValueError):
        list(sanitize(url, None))


@pytest.mark.parametrize(
    "text,expected_column_names",
    (
        ("\"test\"", ("test",)),
        ("\"test\",\"test\"", ("test", "test")),
        ("\"test\", \"test\"", ("test", "test")),
    )
)
def test_parse_column_names(text, expected_column_names):
    assert parse_column_names(text) == expected_column_names


@pytest.mark.parametrize(
    "text,expected_values",
    (
        ("Test", ("Test",)),
        ("Test\tTest", ("Test", "Test")),
        ("Test\tTest\t", ("Test", "Test", "")),
        ("\\N", (None,)),
    )
)
def test_parse_values(text, expected_values):
    assert parse_values(text) == expected_values


@pytest.mark.parametrize('config_type', [
    'no-config', 'empty-config', 'single-column-config'])
@pytest.mark.parametrize('data_label', ['ok', 'invalid'])
def test_optimizations(config_type, data_label):
    if config_type == 'no-config':
        config = None
        decoder_call_count = 0
    else:
        config = Configuration()
        if config_type == 'empty-config':
            decoder_call_count = 0
        else:
            assert config_type == 'single-column-config'
            config.sanitizers["test.notes"] = (lambda x: x)
            decoder_call_count = 3  # Number of rows in test table

    data = {
        'ok': MOCK_PG_DUMP_OUTPUT,
        'invalid': INVALID_MOCK_PG_DUMP_OUTPUT,
    }[data_label]

    should_raise = (
        config_type == 'single-column-config'
        and data_label == 'invalid')

    url = urlparse.urlparse("postgres://localhost/test")
    with mock.patch("subprocess.Popen", side_effect=create_mock_popen(data)):
        with mock.patch.object(dump_postgres, 'decode_copy_value') as decoder:
            decoder.side_effect = decode_copy_value
            if should_raise:
                with pytest.raises(ValueError):
                    list(sanitize(url, config))
            else:
                expected_output = data.decode('utf-8').splitlines()
                assert list(sanitize(url, config)) == expected_output
                assert decoder.call_count == decoder_call_count



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import re

import pymysql
import six


def get_mysqldump_args_and_env_from_url(url):
    """
    Constructs list of command line arguments and dictionary of environment
    variables that can be given to `mysqldump` executable to obtain database
    dump of the database described in given URL.

    :param url: Parsed database URL.
    :type url: urllib.urlparse.ParseResult

    :return: List of command line arguments as well as dictionary of
             environment variables that can be used to launch the MySQL dump
             process to obtain dump of the database.
    :rtype: tuple[list[str],dict[str,str]]
    """
    args = [
        # Without this, `INSERT INTO` statements will exclude column names from
        # the output, which are required for sanitation.
        "--complete-insert",

        # This enables use for "exteded inserts" where multiple rows of a table
        # are included in a single `INSERT INTO` statement (contents of the
        # entire table even, if it's within limits). We use it to increase the
        # performance of the sanitation and to decrease the dump size.
        "--extended-insert",

        # This makes the `mysqldump` to attempt to limit size of a single line
        # into 10 megabytes. We use it to reduce memory consumption.
        "--net_buffer_length=10240",

        # Hostname of the database to connect into, should be always present in
        # the parsed database URL.
        "-h",
        url.hostname,
    ]
    env = {}

    if url.port is not None:
        args.extend(("-P", six.text_type(url.port)))

    if url.username:
        args.extend(("-u", url.username))

    if url.password:
        env["MYSQL_PWD"] = url.password

    if len(url.path) < 2 or not url.path.startswith("/"):
        raise ValueError("Name of the database is missing from the URL")

    args.append(url.path[1:])

    return args, env


MYSQL_NULL_PATTERN = re.compile(r"^NULL$", re.IGNORECASE)
MYSQL_BOOLEAN_PATTERN = re.compile(r"^(TRUE|FALSE)$", re.IGNORECASE)
MYSQL_FLOAT_PATTERN = re.compile(r"^[+-]?\d*\.\d+([eE][+-]?\d+)?$")
MYSQL_INT_PATTERN = re.compile(r"^\d+$")
MYSQL_STRING_PATTERN = re.compile(r"'(?:[^']|''|\\')*(?<![\\])'")


def decode_mysql_literal(text):
    """
    Attempts to decode given MySQL literal into Python value.

    :param text: Value to be decoded, as MySQL literal.
    :type text: str

    :return: Python version of the given MySQL literal.
    :rtype: any
    """
    if MYSQL_NULL_PATTERN.match(text):
        return None

    if MYSQL_BOOLEAN_PATTERN.match(text):
        return text.lower() == "true"

    if MYSQL_FLOAT_PATTERN.match(text):
        return float(text)

    if MYSQL_INT_PATTERN.match(text):
        return int(text)

    if MYSQL_STRING_PATTERN.match(text):
        return decode_mysql_string_literal(text)

    raise ValueError("Unable to decode given value: %r" % (text,))


MYSQL_STRING_ESCAPE_SEQUENCE_PATTERN = re.compile(r"\\(.)")
MYSQL_STRING_ESCAPE_SEQUENCE_MAPPING = {
    "\\0": "\000",
    "\\b": "\b",
    "\\n": "\n",
    "\\r": "\r",
    "\\t": "\t",
    "\\Z": "\032",
}


def decode_mysql_string_literal(text):
    """
    Removes quotes and decodes escape sequences from given MySQL string literal
    returning the result.

    :param text: MySQL string literal, with the quotes still included.
    :type text: str

    :return: Given string literal with quotes removed and escape sequences
             decoded.
    :rtype: str
    """
    assert text.startswith("'")
    assert text.endswith("'")

    # Ditch quotes from the string literal.
    text = text[1:-1]

    return MYSQL_STRING_ESCAPE_SEQUENCE_PATTERN.sub(
        unescape_single_character,
        text,
    )


def unescape_single_character(match):
    """
    Unescape a single escape sequence found from a MySQL string literal,
    according to the rules defined at:
    https://dev.mysql.com/doc/refman/5.6/en/string-literals.html#character-escape-sequences

    :param match: Regular expression match object.

    :return: Unescaped version of given escape sequence.
    :rtype: str
    """
    value = match.group(0)
    assert value.startswith("\\")
    return MYSQL_STRING_ESCAPE_SEQUENCE_MAPPING.get(value) or value[1:]


def encode_mysql_literal(value):
    """
    Converts given Python value into MySQL literal, suitable to be used inside
    `INSERT INTO` statement.

    :param value: Value to convert into MySQL literal.
    :type value: any

    :return: Given value encoded into MySQL literal.
    :rtype: str
    """
    return pymysql.converters.escape_item(value, "utf-8")



def sanitize_null(value):
    return None


def sanitize_empty_json_dict(value):
    return '{}'


def sanitize_empty_json_list(value):
    return '[]'


def sanitize_invalid_django_password(value):
    return '!'



"""
API to sanitation session.

Sanitation session allows having a state within a single sanitation
process.

One important thing stored to the session is a secret key which is
generated to a new random value for each sanitation session, but it
stays constant during the whole sanitation process. Its value is never
revealed, so that it is possible to generate such one way hashes with
it, that should not be redoable afterwards. I.e. during the sanitation
session it's possible to do ``hash(C) -> H`` for any clear text C, but
it is not possible to check if H is the hashed value of C after the
sanitation session has ended.
"""

import hashlib
import hmac
import random
import sys
import threading

from six import int2byte

if sys.version_info >= (3, 6):
    from typing import Callable, Optional, Sequence  # noqa


SECRET_KEY_BITS = 128


_thread_local_storage = threading.local()


def hash_text_to_int(value, bit_length=32):
    # type: (str, int) -> int
    """
    Hash a text value to an integer.

    Generates an integer number based on the hash derived with
    `hash_text` from the given text value.

    :param bit_length: Number of bits to use from the hash value.
    :return: Integer value within ``0 <= result < 2**bit_length``
    """
    hash_value = hash_text(value)
    return int(hash_value[0:(bit_length // 4)], 16)


def hash_text_to_ints(value, bit_lengths=(16, 16, 16, 16)):
    # type: (str, Sequence[int]) -> Sequence[int]
    """
    Hash a text value to a sequence of integers.

    Generates a sequence of integer values with given bit-lengths
    similarly to `hash_text_to_int`, but allowing generating many
    separate numbers with a single call.

    :param bit_lengths:
      Tuple of bit lengths for the resulting integers.  Defines also the
      length of the result tuple.
    :return:
      Tuple of ``n`` integers ``(R_1, ... R_n)`` with the requested
      bit-lengths ``(L_1, ..., L_n)`` and values ranging within
      ``0 <= R_i < 2**L_i`` for each ``i``.
    """
    hash_value = hash_text(value)
    hex_lengths = [x // 4 for x in bit_lengths]
    hex_ranges = (
        (sum(hex_lengths[0:i]), sum(hex_lengths[0:(i + 1)]))
        for i in range(len(hex_lengths)))
    return tuple(int(hash_value[a:b], 16) for (a, b) in hex_ranges)


def hash_text(value, hasher=hashlib.sha256, encoding='utf-8'):
    # type: (str, Callable, str) -> str
    """
    Generate a hash for a text value.

    The hash will be generated by encoding the text to bytes with given
    encoding and then generating a hash with HMAC using the session
    secret as the key and the given hash function.

    :param value: Text value to hash
    :param hasher: Hash function to use, SHA256 by default
    :param encoding: Encoding to use, UTF-8 by default
    :return: Hexadecimal presentation of the hash as a string
    """
    return hash_bytes(value.encode(encoding), hasher)


def hash_bytes(value, hasher=hashlib.sha256):
    # type: (bytes, Callable) -> str
    """
    Generate a hash for a bytes value.

    The hash will be generated by generating a hash with HMAC using the
    session secret as the key and the given hash function.

    :param value: Bytes value to hash
    :param hasher: Hash function to use.
    :return: Hexadecimal presentation of the hash as a string
    """
    return hmac.new(get_secret(), value, hasher).hexdigest()


def get_secret():
    # type: () -> bytes
    """
    Get session specific secret key.

    :return: Session key as bytes
    """
    if not getattr(_thread_local_storage, 'secret_key', None):
        _initialize_session()
    return _thread_local_storage.secret_key  # type: ignore


def reset(secret_key=None):
    # type: (Optional[bytes]) -> None
    """
    Reset the session.

    By default, this resets the value of the secret to None so that, if
    there was an earlier sanitation process ran on the same thread, then
    a next call that needs the secret key of the session will generate a
    new value for it.

    This may also be used to set a predefined value for the secret key.

    :param secret_key:
      Value to set as the new session secret key or None if a new one
      should be generated as soon as one is needed.
    """
    _thread_local_storage.secret_key = secret_key


def _initialize_session():
    # type: () -> None
    """
    Generate a new session key and store it to thread local storage.
    """
    sys_random = random.SystemRandom()
    _thread_local_storage.secret_key = b''.join(
        int2byte(sys_random.randint(0, 255))
        for _ in range(SECRET_KEY_BITS // 8))



from database_sanitizer import session


def setup_module():
    session.reset(b'not-so-secret-key')


def test_hash_text_to_int():
    assert session.hash_text_to_int('hello') == 4100462238


def test_hash_text_to_ints():
    assert session.hash_text_to_ints('hello', [4, 8, 16]) == (15, 70, 33129)


def test_hash_text():
    assert session.hash_text('hello') == (
        'f468169e17f4dd5d7318bd6099a4e657ceb0a978cddb4f3382be0da7121659bb')


def test_hash_bytes():
    assert session.hash_bytes(b'hello') == (
        'f468169e17f4dd5d7318bd6099a4e657ceb0a978cddb4f3382be0da7121659bb')


def test_get_secret():
    assert session.get_secret() == b'not-so-secret-key'


def test_reset():
    old_key = session.get_secret()
    session.reset()
    new_key = session.get_secret()
    assert new_key != old_key
    session.reset(b'not-so-secret-key')
    assert session.get_secret() == b'not-so-secret-key'



import functools
import os

from abc import ABC, abstractmethod

from django.core.management.base import BaseCommand


class WriterBase(ABC):
    @abstractmethod
    def info(self, message: str):
        pass  # pragma: no cover

    @abstractmethod
    def success(self, message: str):
        pass  # pragma: no cover

    @abstractmethod
    def notice(self, message: str):
        pass  # pragma: no cover

    @abstractmethod
    def warning(self, message: str):
        pass  # pragma: no cover

    @abstractmethod
    def error(self, message: str):
        pass  # pragma: no cover


def make_writers(handler: BaseCommand):
    """Make writers.

    Given a BaseCommand instance, returns a dictionary of writer functions
    using the relevant styles.
    """

    class Writer(WriterBase):
        info = functools.partial(
            handler.stdout.write, style_func=handler.style.MIGRATE_HEADING
        )
        success = functools.partial(
            handler.stdout.write, style_func=handler.style.SUCCESS
        )
        notice = functools.partial(
            handler.stdout.write, style_func=handler.style.NOTICE
        )
        warning = functools.partial(
            handler.stdout.write, style_func=handler.style.WARNING
        )
        error = functools.partial(handler.stdout.write, style_func=handler.style.ERROR)

    return Writer()



from database_sanitizer import session
from database_sanitizer.sanitizers import derived


def setup_module():
    session.reset(b'not-so-secret-key')


def test_sanitize_uuid4():
    assert derived.sanitize_uuid4(None) is None
    assert derived.sanitize_uuid4('') == ''
    assert derived.sanitize_uuid4('0') == (
        'e3a5862f-cffb-4d89-ab3e-5563b27e287a')
    assert derived.sanitize_uuid4('00000000000000000000000000000000') == (
        '00000000-0000-0000-0000-000000000000')
    assert derived.sanitize_uuid4('00000000-0000-0000-0000-000000000000') == (
        '00000000-0000-0000-0000-000000000000')
    assert derived.sanitize_uuid4('e3a5862f-cffb-4d89-ab3e-5563b27e287a') == (
        '88b0225e-6090-459a-999d-9b3a3ab28c53')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import mock
import pytest

from ..sanitizers.string import sanitize_empty, sanitize_random, sanitize_zfill


@pytest.mark.parametrize(
    "input_value,expected_output",
    (
        ("foo", ""),
        ("bar", ""),
        ("", ""),
        ("   ", ""),
        (None, None),
    ),
)
def test_sanitize_empty(input_value, expected_output):
    assert sanitize_empty(input_value) == expected_output


@pytest.mark.parametrize(
    "input_value,expected_output",
    (
        ("foo", "000"),
        ("test test", "000000000"),
        ("", ""),
        (None, None)
    ),
)
def test_sanitize_zfill(input_value, expected_output):
    return sanitize_zfill(input_value) == expected_output


@mock.patch('random.choice', return_value='x')
def test_sanitize_random(mocked_random_choice):
    assert sanitize_random(None) is None
    assert sanitize_random('') == ''
    assert sanitize_random('a') == 'x'
    assert sanitize_random('hello') == 'xxxxx'
    assert sanitize_random('hello world') == 'xxxxxxxxxxx'



import subprocess
from io import BytesIO, StringIO

import mock
import pytest

from database_sanitizer import dump
from database_sanitizer.config import Configuration

EXPECTED_POPEN_KWARGS = {
    'mysql://User:Pass@HostName/Db': {
        'args': (
            'mysqldump --complete-insert --extended-insert'
            ' --net_buffer_length=10240 -h hostname -u User Db'
            ' --single-transaction'
        ).split(),
        'env': {'MYSQL_PWD': 'Pass'},
        'stdout': subprocess.PIPE,
    },
    'postgres:///Db': {
        'args': tuple((
            'pg_dump --encoding=utf-8 --quote-all-identifiers'
            ' --dbname postgres:///Db').split()),
        'stdout': subprocess.PIPE,
    },
}

for url in ['postgresql:///Db', 'postgis:///Db']:
    EXPECTED_POPEN_KWARGS[url] = EXPECTED_POPEN_KWARGS['postgres:///Db'].copy()
    EXPECTED_POPEN_KWARGS[url]['args'] = tuple(
        ' '.join(EXPECTED_POPEN_KWARGS[url]['args'])
        .replace('postgres', 'postgresql').split())


@pytest.mark.parametrize('url', list(EXPECTED_POPEN_KWARGS))
@mock.patch('subprocess.Popen')
def test_run(mocked_popen, url):
    mocked_popen.return_value.stdout = BytesIO(b'INPUT DUMP')
    output = StringIO()
    config = None
    dump.run(url, output, config)

    expected_popen_kwargs = EXPECTED_POPEN_KWARGS[url]
    (popen_args, popen_kwargs) = mocked_popen.call_args
    expected_popen_args = (
        (expected_popen_kwargs.pop('args'),) if popen_args else ())
    assert popen_args == expected_popen_args
    assert popen_kwargs == expected_popen_kwargs


@mock.patch('subprocess.Popen')
def test_run_with_mysql_extra_params(mocked_popen):
    mocked_popen.return_value.stdout = BytesIO(b'INPUT DUMP')
    output = StringIO()

    url = "mysql://User:Pass@HostName/Db"
    config = Configuration()
    config.load({
        "config": {
            "extra_parameters": {
                "mysqldump": ["--double-transaction"]
            }
        }
    })

    dump.run(url, output, config)

    expected = {
        'args': (
            'mysqldump --complete-insert --extended-insert'
            ' --net_buffer_length=10240 -h hostname -u User Db'
            ' --double-transaction'
        ).split(),
        'env': {'MYSQL_PWD': 'Pass'},
        'stdout': subprocess.PIPE,
    }

    (popen_args, popen_kwargs) = mocked_popen.call_args
    expected_popen_args = (
        (expected.pop('args'),) if popen_args else ())
    assert popen_args == expected_popen_args
    assert popen_kwargs == expected


@mock.patch('subprocess.Popen')
def test_run_with_pg_dump_extra_params(mocked_popen):
    mocked_popen.return_value.stdout = BytesIO(b'INPUT DUMP')
    output = StringIO()

    url = "postgres:///Db"
    config = Configuration()
    config.load({
        "config": {
            "extra_parameters": {
                "pg_dump": ["--exclude-table=something"]
            }
        }
    })

    dump.run(url, output, config)

    expected = {
        'args': tuple((
            'pg_dump --encoding=utf-8 --quote-all-identifiers'
            ' --dbname postgres:///Db'
            ' --exclude-table=something'
        ).split()),
        'stdout': subprocess.PIPE,
    }

    (popen_args, popen_kwargs) = mocked_popen.call_args
    expected_popen_args = (
        (expected.pop('args'),) if popen_args else ())
    assert popen_args == expected_popen_args
    assert popen_kwargs == expected


@mock.patch('subprocess.Popen')
def test_run_unknown_scheme(mocked_popen):
    with pytest.raises(ValueError) as excinfo:
        dump.run('unknown:///db', None, None)
    assert str(excinfo.value) == "Unsupported database scheme: 'unknown'"
    mocked_popen.assert_not_called()
